----------------------------------------------------------------------------------
-- Engineer:       Armin Diehl
-- Create Date:    22:13:41 01/10/2007 
-- Module Name:    DualPortSram - Behavioral 
-- Target Devices: xc2s200

-- Revision 0.02 - File Created
-- Additional Comments: 
--                 Simple Dual Port interface for static ram
--                 Port1 is read only while port2 is R/W
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

PACKAGE DualPortSram_pckg IS
    COMPONENT DualPortSram IS
    GENERIC(
      FREQ           : natural := 500_000;  -- operating frequency in KHz
      SRAM_SPEED     : natural := 70;       -- ns, can not be less that 20ns
      SRAM_ADDRWIDTH : natural := 15;       -- Ram Address bits
      SRAM_DATAWIDTH : natural := 8
    );
    Port ( reset    : in    STD_LOGIC;      -- high active
           clk      : in    STD_LOGIC;
           -- port1: Read only
           Addr1    : in    STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           DataOut1 : out   STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
           rd1      : in    STD_LOGIC;  -- high active, min 1 clock, max cycles+3
           done1    : out   STD_LOGIC;  -- will be set on falling edge of clock for 1 clock cycle
           -- Port2: Read/Write
           Addr2    : in    STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           DataOut2 : out   STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
           DataIn2  : in    STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
           rd2      : in    STD_LOGIC;
           wr2      : in    STD_LOGIC;
           done2    : out   STD_LOGIC;
	   started2 : out   STD_LOGIC;	-- 1 when access is started, wr/rd2,addr2 and DataIn2 needs to be valid while ^
	   Enable2  : in    STD_LOGIC;  -- can be used to block port2 accesses for some time (from VGA_Sig)
           -- Interface to Static Ram
           SRAddr   : out   STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           SRData   : inout STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
           SR_OE    : out   STD_LOGIC;  -- /OE
           SR_CS    : out   STD_LOGIC;  -- /CS
           SR_WR    : out   STD_LOGIC   -- /WR
    );
   END COMPONENT;
   
END PACKAGE;

----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity DualPortSram is
    generic(
      FREQ           : natural := 100_000;  -- operating frequency in KHz
      SRAM_SPEED     : natural := 70;       -- ns, can not be less that 20ns
      SRAM_ADDRWIDTH : natural := 15;       -- Ram Address bits
      SRAM_DATAWIDTH : natural := 8
    );
    Port ( reset    : in    STD_LOGIC;      -- high active
           clk      : in    STD_LOGIC;
           -- port1: Read only
           Addr1    : in    STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           DataOut1 : out   STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
           rd1      : in    STD_LOGIC;  -- high active, min 1 clock, max cycles+3
           done1    : out   STD_LOGIC;  -- will be set on falling edge of clock for 1 clock cycle
           -- Port2: Read/Write
           Addr2    : in    STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           DataOut2 : out   STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
           DataIn2  : in    STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
           rd2      : in    STD_LOGIC;
           wr2      : in    STD_LOGIC;
           done2    : out   STD_LOGIC;
	   started2 : out   STD_LOGIC;
	   Enable2  : in    STD_LOGIC;  -- can be used to block port2 accesses for some time
           -- Interface to Static Ram
           SRAddr   : out   STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           SRData   : inout STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
           SR_OE    : out   STD_LOGIC;  -- /OE
           SR_CS    : out   STD_LOGIC;  -- /CS
           SR_WR    : out   STD_LOGIC   -- /WR
    );
end DualPortSram;

architecture Behavioral of DualPortSram is
SIGNAL oe_n,cs_n,wr_n : STD_LOGIC;
SIGNAL DataOut : STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
--constant NORM  : natural := 1_000_000;  -- normalize ns * KHz
--constant CYCLES: natural := (SRAM_SPEED*FREQ)/NORM;
constant HI : std_logic := '1';
constant LO : std_logic := '0';
--SIGNAL WaitForRamCounter : INTEGER Range 0 TO CYCLES;
SIGNAL Port1State : INTEGER Range 0 TO 3;
SIGNAL Port2State : INTEGER Range 0 TO 5;
SIGNAL resetAccess,requestPort2,port2Open,Port2Write : STD_LOGIC;
begin
  SR_OE <= oe_n;
  SR_CS <= cs_n;
  SR_WR <= wr_n;
  SRData <= DataOut WHEN Port2Write = HI ELSE (others => 'Z');
  
  gater:PROCESS(Reset,Clk,RequestPort2,Port1State,reset)
  BEGIN
    IF Rising_Edge(Clk) THEN
      IF Reset = '1' THEN
        Port2Open <= '0';
      END IF;
      IF (RequestPort2 = '1') AND (Port1State = 0) THEN
        Port2Open <= '1';
      ELSIF resetAccess = '1' THEN
        Port2Open <= '0';
      END IF;
    END IF;
  END PROCESS;
  
  RamAccess:PROCESS(Reset,Clk,Port2Open,rd1,rd2,wr2)
  BEGIN
    IF Reset = HI THEN
      oe_n <= HI; cs_n <= HI; wr_n <= HI; Port1State <= 0; Port2State <= 0;
      Done1 <= LO; Done2 <= LO;
      requestPort2 <= LO;
      resetAccess <= '0';
      started2 <= LO;
    ELSIF Rising_Edge(Clk) THEN
      -- Port1 reader -----------------------------------------------------------------------
      CASE Port1State IS
        WHEN 0 => IF rd1 = HI THEN
		    Port1State <= 1;
		  END IF;
		  Done1 <= LO;
        WHEN 1 => IF Port2Open = '0' THEN
		    cs_n <= LO; oe_n <= LO;          -- activate ram access
		    SRAddr <= Addr1;
--                    WaitForRamCounter <= CYCLES-1;   -- init wait cycle counter
                    Port1State <= Port1State + 1;
		  END IF;
        WHEN 2 => 
--		  WaitForRamCounter <= WaitForRamCounter - 1;
--                  IF WaitForRamCounter = 0 THEN    -- if we have waited
                    Port1State <= Port1State + 1;
                    DataOut1 <= SRData;            -- get the data
                    Done1 <= HI;                   -- and set done
--                  END IF;
        WHEN 3 => cs_n <= HI; oe_n <= HI;          -- deactivate ram access
                  Port1State <= 0;                 -- end of process
        WHEN others => NULL;
      END CASE;
      
      
      -- Port2 reader/writer -----------------------------------------------------------------
      CASE Port2State IS
        WHEN 0 => IF ((rd2 = HI) OR (wr2 = HI)) AND (Enable2 = HI) THEN
		    Port2State <= 1; 
		    requestPort2 <= HI;
		    Port2Write <= wr2;
		    started2 <= HI;
		  END IF;
		  resetAccess <= '0';
		  Done2 <= LO;
        WHEN 1 => IF Port2Open = '1' THEN
                    SRAddr <= Addr2;			-- let the address stay for one cycle before activating the read
                    Port2State <= Port2State + 1;
		    IF Port2Write = '1' THEN
		      DataOut <= DataIn2;
		    END IF;
                  END IF;
        WHEN 2 => cs_n <= LO;				-- activate ram access
		  wr_n <= not Port2Write;		-- /wr for write
		  oe_n <= Port2Write;			-- /oe only for read
--                  WaitForRamCounter <= CYCLES-1;	-- init wait cycle counter
                  Port2State <= Port2State + 1;
		  started2 <= LO;
        WHEN 3 => 
--		  WaitForRamCounter <= WaitForRamCounter - 1;
--                  IF WaitForRamCounter = 0 THEN		-- if we have waited
                    Port2State <= Port2State + 1;
		    IF Port2Write = LO THEN
                      DataOut2 <= SRData;		-- get the data
		    END IF;
--                  END IF;
        WHEN 4 => cs_n <= HI; oe_n <= HI; wr_n <= HI;	-- deactivate ram access
		  Port2Write <= LO;
                  Port2State <= Port2State + 1;
                  Done2 <= HI;			-- and set done
		  resetAccess <= '1';
	WHEN 5 => requestPort2 <= '0';
		  IF Port2Open = '0' THEN
		    resetAccess <= '0';  
		    Port2State <= 0;
		  END IF;
        WHEN others => null;
      END CASE;
      
    END IF;
  END PROCESS;
  

end Behavioral;

