----------------------------------------------------------------------------------
-- Engineer:		Armin Diehl
-- Create Date:		11:46:08 01/03/2007 
-- Module Name:		VGASignal - Behavioral 
-- Target Devices:	xs2s200
-- Tool versions:	ise 8.2.03i
-- Description: 	Generates VGA signals and signals to fetch pixel data for
--			8 bit font with
-- Revision 0.01 - File Created
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

PACKAGE VGASignal_pckg IS
  COMPONENT VGASignal IS
    Port ( reset           : in  STD_LOGIC;
           Clk25M          : in  STD_LOGIC;
           GetPixelDataSig : out STD_LOGIC;			-- begins 8 pix clocks before first PixClock, will be repeated every 8 clocks, ends before last 8 pixels
           ColCharCount    : out STD_LOGIC_VECTOR (7 downto 0);	-- Counts char in row for fetch (0..255)
           LinePixCount    : out STD_LOGIC_VECTOR (9 downto 0);	-- Line counter (0..479)
           CharPixel       : in  STD_LOGIC_VECTOR (7 downto 0);	-- character pixel data
           ChariRGB        : in  STD_LOGIC_VECTOR (7 downto 0);	-- RGB,Intensity foreground(4 bit) + RGB,Intensity background(4 bit)
           CharReady       : in  STD_LOGIC;			-- on rising edge, CharPixel and IRGBPixel will latched
           hsync_out       : out STD_LOGIC;
           vsync_out       : out STD_LOGIC;
           red_out         : out STD_LOGIC_VECTOR (2 downto 0);
           green_out       : out STD_LOGIC_VECTOR (2 downto 0);
           blue_out        : out STD_LOGIC_VECTOR (2 downto 0);
	   EnableRamAccess : out STD_LOGIC);			-- 1 if access to video ram allowed
   END COMPONENT;
END PACKAGE; 

----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity VGASignal is
    generic(
	HSYNC_START	: natural := 659;
	HSYNC_END	: natural := 755;
	VSYNC_START	: natural := 493;
	VSYNC_END	: natural := 494;
	HCOUNT_MAX	: natural := 799;
	VCOUNT_MAX	: natural := 524;
	XRESOLUTION	: natural := 640;
	YRESOLUTION	: natural := 480
    );
    Port ( reset           : in  STD_LOGIC;
           Clk25M          : in  STD_LOGIC;
           GetPixelDataSig : out STD_LOGIC;			-- begins 8 pix clocks before first PixClock, will be repeated every 8 clocks, ends before last 8 pixels
           ColCharCount    : out STD_LOGIC_VECTOR (7 downto 0);	-- Counts char in row for fetch (0..255)
           LinePixCount    : out STD_LOGIC_VECTOR (9 downto 0);	-- Line counter (0..479)
           CharPixel       : in  STD_LOGIC_VECTOR (7 downto 0);	-- character pixel data
           ChariRGB        : in  STD_LOGIC_VECTOR (7 downto 0);	-- RGB,Intensity foreground(4 bit) + RGB,Intensity background(4 bit)
           CharReady       : in  STD_LOGIC; 			-- on rising edge, CharPixel and IRGBPixel will latched
           hsync_out       : out STD_LOGIC;
           vsync_out       : out STD_LOGIC;
           red_out         : out STD_LOGIC_VECTOR (2 downto 0);
           green_out       : out STD_LOGIC_VECTOR (2 downto 0);
           blue_out        : out STD_LOGIC_VECTOR (2 downto 0);
	   EnableRamAccess : out STD_LOGIC);			-- 1 if access to video ram allowed
end VGASignal;

architecture Behavioral of VGASignal is
SIGNAL horiz_sync, vert_sync : STD_LOGIC;
SIGNAL video_on, video_on_v, video_on_h	: STD_LOGIC;
SIGNAL h_count, v_count			: STD_LOGIC_VECTOR(9 DOWNTO 0);
SIGNAL v_CharCount			: STD_LOGIC_VECTOR(7 DOWNTO 0);	-- col (every 8 pixels) counter for chargen
SIGNAL red,green,blue			: STD_LOGIC_VECTOR(2 DOWNTO 0);
SIGNAL Char,Attr,AttrIn			: STD_LOGIC_VECTOR(7 downto 0);	-- latched from fontgen
SIGNAL CharShift			: STD_LOGIC_VECTOR(7 downto 0);	-- used for shifting char pixel data

begin
-- video_on is high only when RGB data is displayed
video_on <= video_on_H AND video_on_V;


  GenGetPixelDataSig:PROCESS(Clk25M,Reset,h_count,v_count,video_on)
  BEGIN
    IF Reset = '1' THEN
      GetPixelDataSig <= '0';
      v_CharCount <= (others => '0');
    ELSIF Falling_Edge(Clk25M) THEN
	IF (v_count < YRESOLUTION) OR					-- Y on screen
	   ((v_count = VCOUNT_MAX) AND (h_count > HCOUNT_MAX-8)) THEN	-- or last line 
	  IF (h_count = HCOUNT_MAX - 8) THEN
	    GetPixelDataSig <= '1';
	    v_CharCount <= (Others => '0');
	  ELSE
	    IF (h_count(2 downto 0) = "111") AND 
	       ((h_count < XRESOLUTION-9) OR (h_count > HCOUNT_MAX-8)) THEN
	      GetPixelDataSig <= '1';
	      v_CharCount <= v_CharCount + 1;
	    ELSE
	      GetPixelDataSig <= '0';
	    END IF;
	  END IF;
	ELSE
	  GetPixelDataSig <= '0';
	END IF;
    END IF;
  END PROCESS;
  
  genRamAccessEna:PROCESS(Clk25M,Reset,h_count,v_count,video_on)
  BEGIN
    IF Rising_Edge(Clk25M) THEN
      IF Reset = '1' THEN
        EnableRamAccess <= '1';
      ELSIF Video_On = '1' THEN
	EnableRamAccess <= '0';
      ELSIF v_count > YRESOLUTION-1 THEN
	EnableRamAccess <= '1';
      ELSIF h_count < HCOUNT_MAX-24 THEN	-- some spare clocks, GetPixelDataClock begins @ h_count-8
	EnableRamAccess <= '1';
      ELSE
	EnableRamAccess <= '0';
      END IF;
    END IF;
  END PROCESS;
  
  LatchInPixelData:PROCESS(Reset,CharReady)
  BEGIN
    IF Reset = '1' THEN
      Char <= (others => '0');
    ELSIF Rising_Edge(CharReady) THEN
      Char <= CharPixel;
      AttrIn <= ChariRGB;
    END IF;
  END PROCESS;
  
  PixelSifter:PROCESS(Clk25M,h_count,Char)
  BEGIN
    IF Falling_Edge(Clk25M) THEN		-- on falling edge of clock
      IF h_count(2 downto 0) = "111" THEN	-- and end of previous 8 pixel
        CharShift <= Char;			-- get latched char pixel data
	Attr <= AttrIn;				-- and attributes
      ELSE
        CharShift <= 
	  CharShift(6 downto 0) & "0";		-- otherwise shift to next pixel
      END IF;
    END IF;
  END PROCESS;
  
  GenerateRGB:PROCESS(CharShift,Attr)
  BEGIN
    IF Rising_Edge(Clk25M) THEN
      IF CharShift(7) = '1' THEN	-- foreground
        red(2) <= Attr(2);
        green(2) <= Attr(1);
        blue(2) <= Attr(0);
        red(1) <= Attr(3);     		-- indensity
        green(1) <= Attr(3);
        blue(1) <= Attr(3);
        red(0) <= Attr(2);		-- 3 color bits
        green(0) <= Attr(1);
        blue(0) <= Attr(0);
      ELSE
        red(2) <= Attr(6);
        green(2) <= Attr(5);
        blue(2) <= Attr(4);
        red(1) <= Attr(7);		-- indensity (should be blink bit later)
        green(1) <= Attr(7);
        blue(1) <= Attr(7);
        red(0) <= Attr(6);		-- 3 color bits
        green(0) <= Attr(5);
        blue(0) <= Attr(4);
      END IF;
    END IF;
  END Process;

  VClkGen:PROCESS(Reset,Clk25M)
  BEGIN
    if reset = '1' then
      h_count <= "1000001010";   -- pio_bus <= std_logic_vector(to_unsigned(int,pio_bus'LENGTH));
      v_count <= "1000001011";		-- 523
      horiz_sync <= '0';
      vert_sync <= '0';
    elsif rising_edge(Clk25M) then

    -- Generate Horizontal and Vertical Timing Signals for Video Signal
    -- H_count counts pixels (640 + extra time for sync signals)
    -- 
    -- Horiz_sync  ------------------------------------__________--------
    -- H_count     0                640             659       755     799

      IF (h_count = HCOUNT_MAX) THEN
   	h_count <= (others => '0');
	video_on_h <= '1';
      ELSE
        IF (h_count >= XRESOLUTION-1) THEN  -- 639
	  video_on_h <= '0';
	END IF;
   	h_count <= h_count + 1;
      END IF;

    --Generate Horizontal Sync Signal using H_count
      IF (h_count <= HSYNC_END) AND (h_count >= HSYNC_START) THEN
        horiz_sync <= '0';
      ELSE
  	horiz_sync <= '1';
      END IF;

    --V_count counts rows of pixels (480 + extra time for sync signals)
    --  
    --  Vert_sync      -----------------------------------------------_______------------
    --  V_count        0                                      480    493-494          524

      IF (v_count >= VCOUNT_MAX) AND		-- 524
	 (h_count >= HCOUNT_MAX-100) THEN	-- 699
   	v_count <= (Others => '0');
      ELSIF (h_count = HCOUNT_MAX-100) THEN
   	v_count <= v_count + 1;
      END IF;

    -- Generate Vertical Sync Signal using V_count
      IF (v_count <= VSYNC_END) AND (v_count >= VSYNC_START) THEN
	vert_sync <= '0';
      ELSE
	vert_sync <= '1';
      END IF;

      IF (v_count <= YRESOLUTION-1) THEN	-- 479
  	video_on_v <= '1';
	LinePixCount <= v_count;
      ELSE
  	video_on_v <= '0';
	LinePixCount <= (Others => '0');
      END IF;
    
    END IF;
  END PROCESS;
  
-- Put all video signals through DFFs to elminate any delays that cause a blurry image
  red_out <= red when video_on = '1' else "000";
  green_out <= green when video_on = '1' else "000";
  blue_out <= blue when video_on = '1' else "000";
  hsync_out <= horiz_sync;
  vsync_out <= vert_sync;
  ColCharCount <= v_CharCount;
end Behavioral;

