----------------------------------------------------------------------------------
-- Engineer:		Armin Diehl
-- Create Date:		16:54:28 02/11/2007 
-- Module Name:		VGA_Fontgen - Behavioral 
-- Project Name:	VGA Terminal
-- Target Devices:	xc2s200
-- Tool versions:	ise 8.2.03i
-- Description:		Generates char pixel data and cursor
-- Revision 0.01 - File Created
-- 11 Feb 2007 AD:	Inserted one clock to allow font rom data to stabalize
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

PACKAGE VGA_Fontgen_pckg IS
  COMPONENT VGA_Fontgen IS
    GENERIC(
      SRAM_ADDRWIDTH : natural := 15;       -- Ram Address bits
      SRAM_DATAWIDTH : natural := 8
    );
    Port ( reset : in  STD_LOGIC;
           ColCharCount		: in  STD_LOGIC_VECTOR (7 downto 0);
           LinePixCount		: in  STD_LOGIC_VECTOR (9 downto 0);
           clk			: in  STD_LOGIC;
           clkCursorBlink	: in  STD_LOGIC;
	   clkCursorFastBlink	: in  STD_LOGIC;
           clkCharBlink		: in  STD_LOGIC;
           GetPixelDataSig	: in  STD_LOGIC;
           CharPixel		: out  STD_LOGIC_VECTOR (7 downto 0);
           ChariRGB		: out  STD_LOGIC_VECTOR (7 downto 0);
           CharReady		: out  STD_LOGIC;
           CursorVRamAddr	: in  STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           CursorStyle		: in  STD_LOGIC_VECTOR (3 downto 0);
           VideoRamAddr		: inout  STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           VideoRamData		: in  STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
           VideoRamEnable	: out  STD_LOGIC;
           VideoReadDone	: in STD_LOGIC;
           FontRomAddr		: out  STD_LOGIC_VECTOR (11 downto 0);
           FontRomData		: in  STD_LOGIC_VECTOR (7 downto 0)
     );
   END COMPONENT;
END PACKAGE;

----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity VGA_Fontgen is
    GENERIC(
      SRAM_ADDRWIDTH : natural := 15;       -- Ram Address bits
      SRAM_DATAWIDTH : natural := 8
    );
    Port ( reset		: in    STD_LOGIC;
           ColCharCount		: in    STD_LOGIC_VECTOR (7 downto 0);
           LinePixCount		: in    STD_LOGIC_VECTOR (9 downto 0);
           clk			: in    STD_LOGIC;
           clkCursorBlink	: in    STD_LOGIC;
	   clkCursorFastBlink	: in    STD_LOGIC;
           clkCharBlink		: in    STD_LOGIC;
           GetPixelDataSig	: in    STD_LOGIC;
           CharPixel		: out   STD_LOGIC_VECTOR (7 downto 0);
           ChariRGB		: out   STD_LOGIC_VECTOR (7 downto 0);
           CharReady		: out   STD_LOGIC;
           CursorVRamAddr	: in    STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           CursorStyle		: in    STD_LOGIC_VECTOR (3 downto 0);
           VideoRamAddr		: inout STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           VideoRamData		: in    STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
           VideoRamEnable	: out   STD_LOGIC;
           VideoReadDone	: in    STD_LOGIC;
           FontRomAddr		: out   STD_LOGIC_VECTOR (11 downto 0);
           FontRomData		: in    STD_LOGIC_VECTOR (7 downto 0));
end VGA_Fontgen;

architecture Behavioral of VGA_Fontgen is
SIGNAL CharFetchState : Integer Range 0 to 5;
SIGNAL VideoRamAddrLineStart : STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
SIGNAL isCursorPos,cursorOn : STD_LOGIC;
CONSTANT CursorBlinkSlowBit : INTEGER := 0;
CONSTANT CursorBlinkFastBit : INTEGER := 1;
CONSTANT CursorBlockBit     : INTEGER := 2;
CONSTANT CursorLineBit      : INTEGER := 3;
begin

  FetchChar:PROCESS(reset,clk,GetPixelDataSig)
  BEGIN
    IF Reset = '1' THEN
      CharFetchState <= 0;
      CharReady <= '0';
      FontRomAddr <= (Others => '0');
    ELSIF Rising_Edge(Clk) THEN
      CASE CharFetchState IS
        WHEN 0=> CharReady <= '0';
	         IF GetPixelDataSig = '1' THEN
                  CharFetchState <= CharFetchState + 1;		-- start signal for fetch
                  IF (ColCharCount = "00000000") AND
                     (LinePixCount = "0000000000") THEN
                    VideoRamAddr <= (others => '0');		-- start of screen
		    VideoRamAddrLineStart <= (others => '0');
		  ELSIF (ColCharCount = "00000000") AND		-- next character line ?
                        (LinePixCount(3 downto 0) = "000") THEN
		    VideoRamAddrLineStart <= VideoRamAddr;	-- set start address of next line
		  ELSIF (ColCharCount = "00000000") THEN	-- a new line within one character row
		    VideoRamAddr <= VideoRamAddrLineStart;
                  END IF;
                 END IF;

	WHEN 1=> VideoRamEnable <= '1';				-- start read char data from video ram
		 CharFetchState <= CharFetchState + 1;
		 IF (VideoRamAddr = CursorVRamAddr) THEN
		   isCursorPos <= '1';
		 ELSE
		   isCursorPos <= '0';
		 END IF;
		 IF CursorStyle(CursorLineBit downto CursorBlockBit) = "00" THEN
		   cursorOn <= '0';
		 ELSE
		   IF CursorStyle(CursorLineBit) = '1' THEN
		     IF LinePixCount(3 downto 2) = "11" THEN
		       cursorOn <= '1';
		     ELSE
		       cursorOn <= '0';
		     END IF;
		   ELSE
		     cursorOn <= '1';
		   END IF;
		 END IF;

        WHEN 2=> VideoRamEnable <= '0';				-- trigger for read off
		 IF VideoReadDone = '1' THEN			-- wait for Video data ready
								-- set font rom address
		   FontRomAddr <= VideoRamData(15 downto 8) & 
				  LinePixCount(3 downto 0);	-- 16 pixel font height
		   ChariRGB <= VideoRamData(7 downto 0);
		   VideoRamAddr <= VideoRamAddr +1;		-- next char
		   CharFetchState <= CharFetchState + 1;
		 END IF;
        WHEN 3=> CharFetchState <= CharFetchState + 1;		-- time to get font rom data
	WHEN 4=> IF (isCursorPos = '1') AND (cursorOn = '1') THEN
		   IF CursorStyle(CursorBlinkSlowBit) = '1' THEN
		     IF clkCursorBlink = '1' THEN
		       CharPixel <= not FontRomData;
		     ELSE
		       CharPixel <= FontRomData;
		     END IF;
		   ELSIF CursorStyle(CursorBlinkFastBit) = '1' THEN
		     IF clkCursorFastBlink = '1' THEN
		       CharPixel <= not FontRomData;
		     ELSE
		       CharPixel <= FontRomData;
		     END IF;
		   ELSE
		     -- TODO: add non Block cursor
		     CharPixel <= not FontRomData;
		   END IF;
		 ELSE
		   CharPixel <= FontRomData;
		 END IF;
		 
		 CharFetchState <= CharFetchState + 1;

	WHEN 5=> CharReady <= '1';
		 CharFetchState <= 0;
        WHEN others=> CharFetchState <= 0;
      END CASE;
    END IF;
  END PROCESS;

end Behavioral;

